const std  = @import("std");
const dvui = @import("dvui");

const ui = @import("../ui/root.zig");

pub fn gui(src: std.builtin.SourceLocation, ctx: anytype) !void {
    var card = try dvui.box(src, .horizontal, .{
        .id_extra   = ctx.id(),
        .expand     = .horizontal,
        .background = true,
        .border     = dvui.Rect.all(1),
    });
    defer card.deinit();

    {
        var check = ctx.done();
        const changed = try dvui.checkbox(@src(), &check, null, .{
            .gravity_y = 0.5,
            .gravity_x = 0,
        });
        if (changed) {
            std.debug.print("CHANGED\n", .{});
        }
    }

    const hover = true;
    if (hover) {
        var icons = try dvui.box(@src(), .horizontal, .{
            .expand    = .vertical,
            .gravity_x = 1,
        });
        defer icons.deinit();

        if (ctx.staged()) {
            if (try buttonIcon(@src(), "unstage", dvui.entypo.light_down)) {
                std.debug.print("UNSTAGE\n", .{});
            }
        } else {
            if (try buttonIcon(@src(), "stage", dvui.entypo.light_up)) {
                std.debug.print("STAGE\n", .{});
            }
        }
        if (ctx.archived()) {
            if (try buttonIcon(@src(), "unarchive", dvui.entypo.back_in_time)) {
                std.debug.print("UNARCHIVE\n", .{});
            }
        } else {
            if (try buttonIcon(@src(), "archive", dvui.entypo.archive)) {
                std.debug.print("ARCHIVE\n", .{});
            }
        }
    }

    const rc = try title(ctx, card.data().rectScale().r.topLeft());

    // dragging link
    if (!@hasDecl(@TypeOf(ctx.*), "startDragging")) {
        // WIP
        // dvui.pathStroke([_]dvui.Point{rc.topLeft(), cw.});
    }
}

fn buttonIcon(src: std.builtin.SourceLocation, name: []const u8, icon: []const u8) !bool {
    return try dvui.buttonIcon(src, name, icon, .{}, .{
        .margin    = dvui.Rect.all(0),
        .padding   = dvui.Rect.all(4),
        .gravity_y = 0.5,
    });
}

fn title(ctx: anytype, topLeft: dvui.Point) !dvui.Rect {
    // this codes are based on dvui.labelClick

    var lw = dvui.LabelWidget.init(@src(), "#{d} {s}", .{ctx.id(), ctx.summary()}, .{
        .expand    = .horizontal,
        .gravity_x = 0,
        .name      = "LabelClick"
    });
    // now lw has a Rect from its parent but hasn't processed events or drawn

    const lwid = lw.data().id;

    // if lw is visible, we want to be able to keyboard navigate to it
    if (lw.data().visible()) {
        try dvui.tabIndexSet(lwid, lw.data().options.tab_index);
    }

    // draw border and background
    try lw.install();

    // loop over all events this frame in order of arrival
    for (dvui.events()) |*e| {

        // skip if lw would not normally process this event
        if (!lw.matchEvent(e))
            continue;

        switch (e.evt) {
            .mouse => |me| {
                if (me.action == .focus) {
                    e.handled = true;

                    // focus this widget for events after this one (starting with e.num)
                    dvui.focusWidget(lwid, null, e.num);
                } else if (me.action == .press and me.button.pointer()) {
                    e.handled = true;
                    dvui.captureMouse(lwid);

                    // for touch events, we want to cancel our click if a drag is started
                    dvui.dragPreStart(me.p, .{ .name = "TASK", .offset = topLeft.diff(me.p), });
                } else if (me.action == .release and me.button.pointer()) {
                    // mouse button was released, do we still have mouse capture?
                    if (dvui.captured(lwid)) {
                        e.handled = true;

                        // cancel our capture
                        dvui.captureMouse(null);
                        dvui.dragEnd();

                        // if the release was within our border, the click is successful
                        if (lw.data().borderRectScale().r.contains(me.p)) {
                            ctx.open();

                            // if the user interacts successfully with a
                            // widget, it usually means part of the GUI is
                            // changing, so the convention is to call refresh
                            // so the user doesn't have to remember
                            dvui.refresh(null, @src(), lwid);
                        }
                    }
                } else if (me.action == .motion) {
                    if (me.button.touch()) {
                        if (dvui.captured(lwid)) {
                            if (dvui.dragging(me.p)) |_| {
                                // touch: if we overcame the drag threshold, then
                                // that means the person probably didn't want to
                                // touch this button, they were trying to scroll
                                dvui.captureMouse(null);
                                dvui.dragEnd();
                            }
                        }
                    } else if (@hasDecl(@TypeOf(ctx.*), "startDragging")) {
                        // call ctx.startDragging if the method exists
                        if (dvui.captured(lwid)) {
                            e.handled = true;
                            if (dvui.dragging(me.p)) |_| {
                                ctx.startDragging(me.p);
                            }
                        }
                    }
                } else if (me.action == .position) {
                    // a single .position mouse event is at the end of each
                    // frame, so this means the mouse ended above us
                    dvui.cursorSet(.hand);
                }
            },
            .key => |ke| {
                if (ke.action == .down and ke.matchBind("activate")) {
                    e.handled = true;
                    ctx.open();
                    dvui.refresh(null, @src(), lwid);
                }
            },
            else => {},
        }

        // if we didn't handle this event, send it to lw - this means we don't
        // need to call lw.processEvents()
        if (!e.handled) {
            lw.processEvent(e, false);
        }
    }

    // draw text
    try lw.draw();

    // draw an accent border if we are focused
    if (lwid == dvui.focusedWidgetId()) {
        try lw.data().focusBorder();
    }

    // done with lw, have it report min size to parent
    defer lw.deinit();
    return lw.data().scaleRect().r;
}
