//!

const std = @import("std");

///
pub const Self  = @This();

///
pub const Error = error {
    ///
    OutOfMemory,

    ///
    InvalidState,

    ///
    Unknown,
};

///
pub const VTable = struct {
    ///
    deinit: *const fn(ctx: *anyopaque) void,

    ///
    apply: *const fn(ctx: *anyopaque) Error!void,

    ///
    revert: *const fn(ctx: *anyopaque) Error!void,
};

///
pub const Mock = struct {
    target   : ?*i32  = null,
    destroyed: ?*bool = null,

    pub fn init(alloc: std.mem.Allocator, target: ?*i32, destroyed: ?*bool) !Self {
        return make(alloc, Mock {
            .target    = target,
            .destroyed = destroyed,
        });
    }
    pub fn deinit(self: *@This(), _: anytype) void {
        if (self.destroyed) |ptr| {
            ptr.* = true;
        }
    }
    pub fn apply(self: *@This(), _: anytype) !void {
        if (self.target) |ptr| {
            ptr.* += 1;
        }
    }
    pub fn revert(self: *@This(), _: anytype) !void {
        if (self.target) |ptr| {
            ptr.* -= 1;
        }
    }
};

///
ctx: *anyopaque,

///
vtable: *const VTable,

///
pub fn deinit(self: Self) void {
    self.vtable.deinit(self.ctx);
}

///
pub fn apply(self: Self) !void {
    return self.vtable.apply(self.ctx);
}

///
pub fn revert(self: Self) !void {
    return self.vtable.revert(self.ctx);
}

///
pub fn make(alloc: std.mem.Allocator, ctx: anytype) !Self {
    const Target = @TypeOf(ctx);
    const Wrapper = struct {
        const vtable = VTable {
            .deinit = @This().del,
            .apply  = @This().apply,
            .revert = @This().revert,
        };

        alloc: std.mem.Allocator,
        ctx  : Target,

        fn new(alloc2: std.mem.Allocator, ctx2: Target) !*@This() {
            const self = try alloc2.create(@This());
            self.* = .{
                .alloc = alloc2,
                .ctx   = ctx2,
            };
            return self;
        }
        fn del(selfp: *anyopaque) void {
            const self: *@This() = @ptrCast(@alignCast(selfp));
            self.ctx.deinit(self.alloc);
            self.alloc.destroy(self);
        }
        fn apply(selfp: *anyopaque) Error!void {
            const self: *@This() = @ptrCast(@alignCast(selfp));
            return self.ctx.apply(self.alloc);
        }
        fn revert(selfp: *anyopaque) Error!void {
            const self: *@This() = @ptrCast(@alignCast(selfp));
            return self.ctx.revert(self.alloc);
        }
    };
    return Self {
        .ctx    = try Wrapper.new(alloc, ctx),
        .vtable = &Wrapper.vtable,
    };
}

test {
    var value: i32 = 0;
    var destroyed = false;
    {
        const cmd = try Mock.init(std.testing.allocator, &value, &destroyed);
        defer cmd.deinit();

        try cmd.apply();
        try std.testing.expectEqual(value, 1);

        try cmd.revert();
        try std.testing.expectEqual(value, 0);
    }

    try std.testing.expect(destroyed);
}
